;; -*- coding: utf-8; mode: scheme -*-
;;
;; avatar.scm
;;
;;   Copyright (c) 2008 KOGURO, Naoki (naoki@koguro.net)
;;   All rights reserved.
;;
;;   Redistribution and use in source and binary forms, with or without 
;;   modification, are permitted provided that the following conditions 
;;   are met:
;;
;;   1. Redistributions of source code must retain the above copyright 
;;      notice, this list of conditions and the following disclaimer.
;;   2. Redistributions in binary form must reproduce the above copyright 
;;      notice, this list of conditions and the following disclaimer in the 
;;      documentation and/or other materials provided with the distribution.
;;   3. Neither the name of the authors nor the names of its contributors 
;;      may be used to endorse or promote products derived from this 
;;      software without specific prior written permission.
;;
;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
;;   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
;;   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
;;   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
;;   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
;;   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;

(use gl)
(use gl.glut)
(use artoolkit)
(use gauche.sequence)
(use srfi-13)
(use util.match)
(use object)
(use gauche.uvector)
(use gauche.charconv)
(use gauche.net)
(use gauche.selector)
(use util.queue)
(use gauche.parseopt)

(load "fonts16.scm")

(define patt-id #f)

(define text-buffer (make-u8vector 131072 0))
(define text-texture #f)

(define selector #f)

(define show-avatar? #f)
(define show-text? #f)
(define left-arm-angle 45)
(define right-arm-angle 45)
(define head-angle 0)

(define (char-code->iso2022jp code)
  (with-input-from-string (ces-convert (string (integer->char code))
                                       (gauche-character-encoding)
                                       'iso2022jp)
    (lambda ()
      (let ((v 0))
        (letrec ((normal (lambda (b)
                           (case b
                             ((27) escape0)
                             (else
                              (set! v (+ b (* v 256)))
                              normal))))
                 (escape0 (lambda (b)
                            escape1))
                 (escape1 (lambda (b)
                            normal)))
          (port-fold (lambda (b f)
                       (f b))
                     normal
                     read-byte)
          v)))))

(define (fb-draw-char! x y code)
  (match-let (((width . pixel-list) (vector-ref fonts16
                                                (char-code->iso2022jp code))))
      (for-each (match-lambda ((dx dy)
                               (let ((i (* (+ (+ x dx) (* 256 (+ y dy))) 4)))
                                 (u8vector-set! text-buffer (+ i 0) 255)
                                 (u8vector-set! text-buffer (+ i 1) 255)
                                 (u8vector-set! text-buffer (+ i 2) 255)
                                 (u8vector-set! text-buffer (+ i 3) 255))))
                pixel-list)
    width))

(define (fb-draw-string! x y str)
  (fold (lambda (char px)
          (+ px (fb-draw-char! px y (char->integer char))))
        x
        (string->list str)))

(define command-queue (make-queue))

(define (init)
  (ar-video-open "")
  (receive (xsize ysize) (ar-video-inq-size)
    (format #t "Image size (x, y) = (~d, ~d)~%" xsize ysize)
    (let ((cparam (ar-param-change-size (ref (ar-param-load "camera_para.dat" 1) 0)
                                        xsize ysize)))
      (ar-init-cparam cparam)
      (ar-param-disp cparam)

      (set! patt-id (ar-load-patt "./gauche.patt"))

      (arg-init cparam 1.0 0 0 0 #f)))

  (load-material "./sample.mtl")
  (load-object "./sample.obj")

  (set! text-texture (ref (gl-gen-textures 1) 0))

  (set! selector (make <selector>))
  (let ((server (make-server-socket 'inet 12345 :reuse-addr? #t)))
    (define (accept-handler sock flag)
      (let* ((client (socket-accept server))
             (output (socket-output-port client)))
        (selector-add! selector (socket-input-port client :buffering #f)
                       (lambda (input flag)
                         (read-command client input output))
                       '(r))))
    (define (read-command client input output)
      (let ((sexp (read input)))
        (if (eof-object? sexp)
            (begin
              (selector-delete! selector input #f #f)
              (socket-close client))
            (enqueue! command-queue sexp))))
    (selector-add! selector (socket-fd server) accept-handler '(r)))
  )

(define (eval-command sexp)
  (match sexp
    (('show-avatar)
     (set! show-avatar? #t))
    (('hide-avatar)
     (set! show-avatar? #f))
    (('show-text)
     (set! show-text? #t))
    (('hide-text)
     (set! show-text? #f))
    (('clear-text)
     (u8vector-fill! text-buffer 0))
    (('say n text)
     (fb-draw-string! 0 (* n 16) text))
    (('right-arm n)
     (set! right-arm-angle n))
    (('left-arm n)
     (set! left-arm-angle n))
    (('head n)
     (set! head-angle n))
    (else
     #f)))

(define (draw patt-trans)
  (arg-draw-mode-3d)
  (arg-draw-3d-camera 0 0)
  (gl-clear-depth 1.0)
  (gl-clear GL_DEPTH_BUFFER_BIT)
  (gl-enable GL_DEPTH_TEST)
  (gl-depth-func GL_LEQUAL)

  (selector-select selector 0)
  (for-each eval-command (dequeue-all! command-queue))
  
  (gl-matrix-mode GL_MODELVIEW)
  (gl-load-matrix (arg-conv-glpara patt-trans))

  (gl-enable GL_LIGHTING)
  (gl-enable GL_LIGHT0)
  (gl-light GL_LIGHT0 GL_POSITION #f32(100.0 -200.0 200.0 0.0))
  (gl-light GL_LIGHT0 GL_AMBIENT #f32(0.1 0.1 0.1 0.1))
  (gl-light GL_LIGHT0 GL_DIFFUSE #f32(0.9 0.9 0.9 0.1))
  
  (when show-avatar?
    (gl-push-matrix)
    (gl-rotate 90.0 1.0 0.0 0.0)
    (gl-scale 3.0 3.0 3.0)
    (gl-translate 0 30 0)

    (gl-push-matrix)
    (gl-rotate head-angle 0 1 0)
    (draw-object 'ungrouped)
    (draw-object 'face)
    (draw-object 'head_back)
    (draw-object 'hat)
    (gl-pop-matrix)
    
    (draw-object 'body)
    (draw-object 'pants)
    (draw-object 'shoes)
    (draw-object 'back)

    (gl-push-matrix)
    (gl-translate -8.3 1 0)
    (gl-rotate right-arm-angle 0 0 1)
    (gl-translate 8.3 -1 0)
    (draw-object 'left_arm)
    (draw-object 'left_hand)
    (gl-pop-matrix)

    (gl-push-matrix)
    (gl-translate 8.3 1 0)
    (gl-rotate left-arm-angle 0 0 -1)
    (gl-translate -8.3 -1 0)
    (draw-object 'right_arm)
    (draw-object 'right_hand)
    (gl-pop-matrix)

    (when show-text?
      (gl-enable GL_TEXTURE_2D)
      (gl-material GL_FRONT_AND_BACK GL_DIFFUSE #f32(0 1 0.3 0.7))
      (gl-bind-texture GL_TEXTURE_2D text-texture)
      (gl-tex-image-2d GL_TEXTURE_2D 0 4 256 128 0 GL_RGBA GL_UNSIGNED_BYTE text-buffer)
      (gl-tex-parameter GL_TEXTURE_2D GL_TEXTURE_MIN_FILTER GL_LINEAR)
      (gl-begin GL_POLYGON)
      (gl-tex-coord 1 1)
      (gl-vertex -20 15)
      (gl-tex-coord 1 0)
      (gl-vertex -20 80)
      (gl-tex-coord 0 0)
      (gl-vertex -150 80)
      (gl-tex-coord 0 1)
      (gl-vertex -150 15)
      (gl-end))

    (gl-pop-matrix))

  (gl-disable GL_LIGHTING)
  (gl-disable GL_DEPTH_TEST))

(define (main-loop)
  (or (and-let* ((data (ar-video-get-image)))
        (arg-draw-mode-2d)
        (arg-disp-image data 0 0)
        (let ((markers (ar-detect-marker data 100)))
          (ar-video-cap-next)
          (and-let* ((marker (find (lambda (m)
                                     (eq? patt-id (slot-ref m 'id)))
                                   markers)))
            (draw (ar-get-trans-mat marker #f64(0.0 0.0) 80.0)))
          (arg-swap-buffers)
          #t))
      (ar-util-sleep 2)))

(define (main args)
  (glut-init '())
  (init)

  (ar-video-cap-start)

  (arg-main-loop #f
                 (lambda (key x y)
                   (when (= key #x1b)
                     (ar-video-cap-stop)
                     (ar-video-close)
                     (arg-cleanup)
                     (exit 0)))
                 main-loop)

  0)
  
;; end of file

