;; -*- coding: utf-8; mode: scheme -*-
;;
;; object.scm 
;;
;;   Copyright (c) 2008 KOGURO, Naoki (naoki@koguro.net)
;;   All rights reserved.
;;
;;   Redistribution and use in source and binary forms, with or without 
;;   modification, are permitted provided that the following conditions 
;;   are met:
;;
;;   1. Redistributions of source code must retain the above copyright 
;;      notice, this list of conditions and the following disclaimer.
;;   2. Redistributions in binary form must reproduce the above copyright 
;;      notice, this list of conditions and the following disclaimer in the 
;;      documentation and/or other materials provided with the distribution.
;;   3. Neither the name of the authors nor the names of its contributors 
;;      may be used to endorse or promote products derived from this 
;;      software without specific prior written permission.
;;
;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
;;   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
;;   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
;;   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
;;   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
;;   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;

(define-module object
  (use gl)
  (use gl.glut)
  (use gl.simple-image)
  (use gauche.uvector)
  (use util.match)

  (export-all)
  )

(select-module object)

(define-class <material> ()
  ((ambient :init-value #f
            :accessor ambient-of)
   (diffuse :init-value #f
            :accessor diffuse-of)
   (specular :init-value #f
             :accessor specular-of)
   (texture-id :init-value #f
               :accessor texture-id-of)))

(define-class <vertex> ()
  ((point :init-value #f
          :accessor point-of)
   (texture :init-value #f
            :accessor texture-of)
   (normal :init-value #f
           :accessor normal-of)))

(define-class <polygon> ()
  ((vertex-list :init-form (list)
                :accessor vertex-list-of
                :init-keyword :vertex-list)))

(define-class <group> ()
  ((material :init-value #f
             :accessor material-of)
   (polygon-list :init-form (list)
                 :accessor polygon-list-of)))
   
(define material-table (make-hash-table 'eq?))

(define (find-material name)
  (hash-table-get material-table name #f))

(define (load-material filename)
  (with-input-from-file filename
    (lambda ()
      (let ((material #f))
        (port-for-each (lambda (line)
                         (rxmatch-case line
                           (#/newmtl (.*)/ (#f name)
                            (set! material (make <material>))
                            (hash-table-put! material-table
                                             (string->symbol name)
                                             material))
                           (#/Kd ([\d.\-]+) ([\d.\-]+) ([\d.\-]+)/ (#f r g b)
                            (when material
                              (set! (diffuse-of material)
                                    (f32vector (x->number r)
                                               (x->number g)
                                               (x->number b)
                                               1.0))))
                           (#/Ks ([\d.\-]+) ([\d.\-]+) ([\d.\-]+)/ (#f r g b)
                            (when material
                              (set! (specular-of material)
                                    (f32vector (x->number r)
                                               (x->number g)
                                               (x->number b)
                                               1.0))))
                           (#/Ka ([\d.\-]+) ([\d.\-]+) ([\d.\-]+)/ (#f r g b)
                            (when material
                              (set! (ambient-of material)
                                    (f32vector (x->number r)
                                               (x->number g)
                                               (x->number b)
                                               1.0))))
                           (#/map_Kd (.*)/ (#f filename)
                            (match-let1 (width height depth image)
                                (read-sgi-image filename)
                              (let* ((textures (gl-gen-textures 1))
                                     (texture (ref textures 0)))
                                (gl-bind-texture GL_TEXTURE_2D texture)
                                (gl-tex-image-2d GL_TEXTURE_2D 0 depth width height 0 GL_RGB GL_UNSIGNED_BYTE image)
                                (gl-tex-parameter GL_TEXTURE_2D GL_TEXTURE_MIN_FILTER GL_LINEAR)
                                (set! (texture-id-of material) texture))))
                           (else
                            #f)))
                       read-line)))))

(define group-table (make-hash-table 'eq?))

(define (find-group name)
  (hash-table-get group-table name #f))

(define (load-object filename)
  (define v-table (make-hash-table 'equal?))
  (define vt-table (make-hash-table 'equal?))
  (define vn-table (make-hash-table 'equal?))
  (define v-index 1)
  (define vt-index 1)
  (define vn-index 1)

  (with-input-from-file filename
    (lambda ()
      (let ((group #f)
            (material #f))
        (port-for-each (lambda (line)
                         (rxmatch-case line
                           (#/g (.*)/ (#f name)
                            (set! group (make <group>))
                            (set! (material-of group) material)
                            (hash-table-put! group-table
                                             (string->symbol name)
                                             group))
                           (#/usemtl (.*)/ (#f name)
                            (set! material
                                  (find-material (string->symbol name))))
                           (#/v ([\d.\-]+) ([\d.\-]+) ([\d.\-]+)/ (#f x y z)
                            (hash-table-put! v-table v-index
                                             (list->point4f (map x->number
                                                                 (list x y z))))
                            (inc! v-index))
                           (#/vt ([\d.\-]+) ([\d.\-]+)/ (#f s t)
                            (hash-table-put! vt-table vt-index
                                             (f32vector (x->number s)
                                                        (x->number t)))
                            (inc! vt-index))
                           (#/vn ([\d.\-]+) ([\d.\-]+) ([\d.\-]+)/ (#f x y z)
                            (hash-table-put! vn-table vn-index
                                             (list->vector4f (map x->number
                                                                  (list x y z))))
                            (inc! vn-index))
                           (#/f (.*)/ (#f args)
                            (set! (polygon-list-of group)
                                  (append (polygon-list-of group)
                                          (list (make <polygon>
                                                  :vertex-list
                                                  (map (lambda (str)
                                                         (let ((lst (map x->number (string-split str "/")))
                                                               (v (make <vertex>)))
                                                           (set! (point-of v)
                                                                 (hash-table-get v-table (ref lst 0)))
                                                           (set! (texture-of v)
                                                                 (hash-table-get vt-table (ref lst 1)))
                                                           (set! (normal-of v)
                                                                 (hash-table-get vn-table (ref lst 2)))
                                                           v))
                                                       (string-split args " ")))))))
                           (else
                            #f)))
                       read-line)))))

(define (draw-object name)
  (let ((group (find-group name)))
    (when group
      (draw-group group))))

(define (draw-group group)
  (draw-material (material-of group))
  (for-each (lambda (polygon)
              (gl-begin GL_POLYGON)
              (for-each (lambda (v)
                          (when (and (material-of group)
                                     (texture-id-of (material-of group)))
                            (gl-tex-coord (texture-of v)))
                          (gl-normal (normal-of v))
                          (gl-vertex (point-of v)))
                        (vertex-list-of polygon))
              (gl-end)
              )
            (polygon-list-of group)))

(define (draw-material material)
  (if material
      (begin
        (gl-material GL_FRONT GL_AMBIENT (ambient-of material))
        (gl-material GL_FRONT GL_DIFFUSE (diffuse-of material))
        (gl-material GL_FRONT GL_SPECULAR (specular-of material))
        (if (texture-id-of material)
            (begin
              (gl-enable GL_TEXTURE_2D)
              (gl-bind-texture GL_TEXTURE_2D (texture-id-of material)))
            (gl-disable GL_TEXTURE_2D)))
      (gl-disable GL_TEXTURE_2D)
    ))

(define (draw-all-groups)
  (for-each draw-group (hash-table-values group-table)))

(provide "object")


;; end of file
